/* 
Copyright (C) 2010 Kevin Allen

This file is part of kacq.

kacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kacq.  If not, see <http://www.gnu.org/licenses/>.



*/
#include <gtk/gtk.h>
#include "main.h"

int init_window()
{
  // function to complete the interface built with glade
  builder = gtk_builder_new ();
  if(gtk_builder_add_from_file (builder, "gacq.glade", NULL)==0)
    {
      fprintf(stderr,"An error occurred reading the glade file\n" );
      return -1;
    }
  // get a reference for the widget we need to play with
  widgets.window = GTK_WIDGET (gtk_builder_get_object (builder, "window"));
  widgets.vbox1 = GTK_WIDGET (gtk_builder_get_object (builder, "vbox1"));
  widgets.about_dlg= GTK_WIDGET (gtk_builder_get_object (builder, "about_dialog"));
  widgets.preferences_dlg= GTK_WIDGET (gtk_builder_get_object (builder, "preferences_dialog"));
  widgets.toolbar= GTK_WIDGET (gtk_builder_get_object (builder, "toolbar"));
  widgets.dev1_name_label= GTK_WIDGET (gtk_builder_get_object (builder, "device_1_name_label"));
  widgets.dev2_name_label = GTK_WIDGET (gtk_builder_get_object (builder, "device_2_name_label"));
  widgets.dev1_driver_label= GTK_WIDGET (gtk_builder_get_object (builder, "device_1_driver_label"));
  widgets.dev2_driver_label= GTK_WIDGET (gtk_builder_get_object (builder, "device_2_driver_label"));
  widgets.num_devices_detected_label= GTK_WIDGET (gtk_builder_get_object (builder, "num_device_detected_label"));
  widgets.num_available_channels_label=GTK_WIDGET (gtk_builder_get_object (builder, "num_available_channels_label"));
  widgets.sampling_rate_entry= GTK_WIDGET (gtk_builder_get_object (builder, "sampling_rate_entry"));
  widgets.num_channels_device_1_label=GTK_WIDGET (gtk_builder_get_object (builder, "num_channels_device_1_label"));
  widgets.num_channels_device_2_label=GTK_WIDGET (gtk_builder_get_object (builder, "num_channels_device_2_label"));
  widgets.preferences_channel_vbox=GTK_WIDGET (gtk_builder_get_object (builder, "preferences_channel_vbox"));
  widgets.file_name_entry=GTK_WIDGET(gtk_builder_get_object (builder, "file_name_entry"));
  widgets.trial_spinbutton=GTK_WIDGET(gtk_builder_get_object (builder, "trial_spinbutton"));
  widgets.statusbar=GTK_WIDGET(gtk_builder_get_object (builder, "statusbar"));
  widgets.preferences_channel_vbox=GTK_WIDGET(gtk_builder_get_object (builder, "preferences_channel_vbox"));
  widgets.channel_view=GTK_WIDGET(gtk_builder_get_object (builder, "view"));




  
  // put a gtkdatabox in the vbox
  // Create the GtkDatabox widget
  gtk_databox_create_box_with_scrollbars_and_rulers(&widgets.box, &widgets.table,TRUE, TRUE, TRUE, TRUE);
  // Put it in the vbox1
  gtk_box_pack_start (GTK_BOX (widgets.vbox1), widgets.table, TRUE, TRUE, 0);
  int num_points=100000;
  gfloat *X;
  gfloat *Y;
  int i;
  // We define some data 
  X = g_new0 (gfloat, num_points);
  Y = g_new0 (gfloat, num_points);
  
  for (i = 0; i < num_points; i++)
    {
      X[i] = i;
      Y[i] = sin (i * 4 * G_PI / num_points);
    }
  
  widgets.color.red = 0;
  widgets.color.green = 0;
  widgets.color.blue = 0;
  widgets.graph = gtk_databox_points_new (num_points, X, Y, &widgets.color, 1);
  gtk_databox_graph_add (GTK_DATABOX (widgets.box), widgets.graph);
  gtk_databox_set_total_limits (GTK_DATABOX (widgets.box), -1000., 5000., -10000., 23000.);
  gtk_databox_auto_rescale (GTK_DATABOX (widgets.box), 0.05);
  
  //  widgets.test_label=gtk_label_new("view");
  // gtk_box_pack_start (GTK_BOX(widgets.vbox1), widgets.test_label, FALSE,FALSE, 0);
  gtk_box_reorder_child(GTK_BOX(widgets.vbox1),widgets.table,1);
  gtk_widget_show (widgets.table);
  gtk_widget_show (widgets.box);


  // so that comedi page of preferences dialog show the right labels
  set_device_labels();

  // so that the channel page of the preferences dialog show the right labels
  build_channel_tree_view();
  
  gtk_builder_connect_signals (builder, NULL);          
  g_object_unref (G_OBJECT (builder));
  
  // show the main window
  gtk_widget_show (widgets.window);      
  return 0;
}
void create_and_fill_model()
{
  GtkTreeIter    iter;
  int i, j, channel_index;
  widgets.channel_store = gtk_list_store_new (NUM_COLS, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
  channel_index=0;
  for (i=0; i < comedi_inter.number_devices;i++)
    {
      for (j=0; j < comedi_inter.dev[i].number_channels_analog_input; j++)
	{
	  // by default, all the channels are selected
	  gtk_list_store_append (widgets.channel_store, &iter);
	  gtk_list_store_set (widgets.channel_store, &iter,
			      COL_NO, channel_index,
			      COL_DEVICE, i,
			      COL_SELECT, TRUE,
			      -1);
	  channel_index++; // increase the index for next channel
	}
    }
  return;
}
void cell_toggled_callback (GtkCellRendererToggle *cell,
			    gchar                 *path_string,
			    gpointer               user_data)
{
  // function to select and deselect channels in preferences dialog
  GtkTreeIter iter;
  gboolean enabled;
  GtkTreePath* path = gtk_tree_path_new_from_string(path_string);
  gtk_tree_model_get_iter(GTK_TREE_MODEL (widgets.channel_store), &iter, path);
  gtk_tree_model_get(GTK_TREE_MODEL(widgets.channel_store), &iter, COL_SELECT, &enabled, -1);
  enabled = !enabled;
  gtk_list_store_set(widgets.channel_store, &iter, COL_SELECT, enabled, -1);
  return;
}

int build_channel_tree_view()
{
  GtkCellRenderer     *renderer;
  
  /* --- Column #1 --- */
  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widgets.channel_view),
                                               -1,      
                                               "No",  
                                               renderer,
                                               "text", COL_NO,
                                               NULL);
  /* --- Column #2 --- */
  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widgets.channel_view),
                                               -1,      
                                               "Device",  
                                               renderer,
                                               "text", COL_DEVICE,
                                               NULL);
  /* --- Column #3 --- */
  renderer=gtk_cell_renderer_toggle_new();
  g_object_set(renderer,"active", TRUE, NULL);
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widgets.channel_view),
					       -1,      
					       "Selected",  
					       renderer,
					       "active", COL_SELECT,
                                                 NULL );
  g_signal_connect(renderer, "toggled", (GCallback) cell_toggled_callback, NULL);
  create_and_fill_model ();
  gtk_tree_view_set_model (GTK_TREE_VIEW (widgets.channel_view),GTK_TREE_MODEL(widgets.channel_store));
  return 0;
}
int set_device_labels()
{
  gchar *str;
  str=g_strdup_printf("%d", comedi_inter.sampling_rate);
  gtk_entry_set_text(GTK_ENTRY(widgets.sampling_rate_entry),str);
  str=g_strdup_printf("%d", comedi_inter.number_channels);
  gtk_label_set_text(GTK_LABEL(widgets.num_available_channels_label),str);


  if(comedi_inter.number_devices>0)
    {
      str=g_strdup_printf("%d", comedi_inter.number_devices);
      gtk_label_set_text(GTK_LABEL(widgets.num_devices_detected_label),str);
      gtk_label_set_text(GTK_LABEL(widgets.dev1_name_label),comedi_inter.dev[0].name);
      gtk_label_set_text(GTK_LABEL(widgets.dev1_driver_label),comedi_inter.dev[0].driver);
      str=g_strdup_printf("%d", comedi_inter.dev[0].number_channels_analog_input);
      gtk_label_set_text(GTK_LABEL(widgets.num_channels_device_1_label),str);
    }
  if(comedi_inter.number_devices>1)
    {
      gtk_label_set_text(GTK_LABEL(widgets.dev2_name_label),comedi_inter.dev[1].name);
      gtk_label_set_text(GTK_LABEL(widgets.dev2_driver_label),comedi_inter.dev[1].driver);
      str=g_strdup_printf("%d", comedi_inter.dev[1].number_channels_analog_input);
      gtk_label_set_text(GTK_LABEL(widgets.num_channels_device_2_label),str);
    }
  else
    {
      gtk_label_set_text(GTK_LABEL(widgets.dev2_name_label),"None");
      gtk_label_set_text(GTK_LABEL(widgets.dev2_driver_label),"None");
    }
  g_free(str);
  return 0;
}

// callback for the main window 
void on_window_destroy (GtkObject *object, gpointer user_data)
{
  gtk_main_quit();
} 

// callback for the about dialog
void on_about_dialog_delete_event(GtkObject *object, gpointer user_data)
{
  gtk_widget_hide(widgets.about_dlg);
}
// callback for the preferences dialog
void on_preferences_dialog_delete_event(GtkObject *object, gpointer user_data)
{
  gtk_widget_hide(widgets.preferences_dlg);
}

// callback for the menu item
void on_open_menuitem_activate(GtkObject *object, gpointer user_data)
{
   fprintf(stderr,"on open menuitem activate\n");
}
void on_save_menuitem_activate(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"on save menuitem activate\n");
}
void on_save_as_menuitem_activate(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"on save as menuitem activate\n");
}
void on_quit_menuitem_activate(GtkObject *object, gpointer user_data)
{
  gtk_main_quit();
} 
void on_about_menuitem_activate(GtkObject *object, gpointer user_data)
{
  int response;
  response=gtk_dialog_run(GTK_DIALOG(widgets.about_dlg));
  if ( response ==GTK_RESPONSE_CANCEL)
    {
      gtk_widget_hide(widgets.about_dlg);
    }
}
void on_preferences_menuitem_activate(GtkObject *object, gpointer user_data)
{
  int response;
  response=gtk_dialog_run(GTK_DIALOG(widgets.preferences_dlg));
  if ( response ==GTK_RESPONSE_CANCEL)
    {
      fprintf(stderr,"cancel preferences dlg\n");
      gtk_widget_hide(widgets.preferences_dlg);
    }
    if ( response ==GTK_RESPONSE_OK)
    {
      fprintf(stderr,"ok preferences dlg\n");
      gtk_widget_hide(widgets.preferences_dlg);
    }
}
void on_ok_preferences_button_activate(GtkObject *object, gpointer user_data)
{
  gtk_widget_hide(widgets.preferences_dlg);
  fprintf(stderr,"hide the preference dlg\n");
}

void on_previous_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
   fprintf(stderr,"previous button\n");
}
void on_rewind_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
   fprintf(stderr,"rewind button\n");
}
void on_play_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"play button\n");
  if(comedi_inter.is_acquiring==0)
    {
      // if thread is started successfully, set the flag to 1
      if( (acquisition_thread_id=pthread_create(&acquisition_thread, NULL, acquisition, (void*)t))==1)
	{
	  fprintf(stderr,"error creating the acquisition thread, error_no:%d\n",
		  acquisition_thread_id);
	  return;
	}
      comedi_inter.is_acquiring=1;
    }
}
void on_stop_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"should stop recording and acquisition threads\n");
  comedi_inter.is_acquiring=0;
  recording_inter.is_recording=0;
  // sleep a minimum of time so the 2 threads die
  nanosleep(&comedi_inter.inter_acquisition_sleep_timespec,&comedi_inter.req);
  fprintf(stderr,"stop button\n");
}
void on_forward_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"forward button\n");
}
void on_next_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"next button\n");
}
void on_record_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  const gchar *str;
  gchar * str1;
  gchar * str2;
  gchar * str3;

  // if already recording, do nothing and exit here
  if (recording_inter.is_recording==1)
    {
      fprintf(stderr,"record button, is already recording, this is a but\n");
      return ;
    }
  // if acquisition thread is already running, kill it
  if(comedi_inter.is_acquiring==1)
    {
      fprintf(stderr,"on_record_toolbutton_clicked, need to kill acquisition thread\n");
      comedi_inter.is_acquiring=0;
      // sleep a few ms to allow the old thread to complete and die
      nanosleep(&comedi_inter.timespec_pause_restat_acquisition_thread,&comedi_inter.req);
    }

  
  // update the list of channels to record
  recording_interface_get_channels_from_channel_store(&recording_inter,&comedi_inter,widgets.channel_store);

  // get the file name from the gui
  str=gtk_entry_get_text(GTK_ENTRY(widgets.file_name_entry));
  str1=g_strdup_printf("%02d",gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widgets.trial_spinbutton)));
  str2=".dat";
  str3=g_strdup_printf("%s_%s%s",str,str1,str2);
  recording_inter.file_name=g_strdup_printf("%s%s",recording_inter.directory,str3);
    
  
  if ((recording_interface_open_file(&recording_inter))!=0)
    {
      fprintf(stderr,"recording(): unable to open file, recording aborted\n");
      recording_inter.is_recording=0;
      return ;
    }



  // set the statusbar to recording details
 
  str=g_strdup_printf("Recording from %d channels at %d Hz, saved in %s",recording_inter.number_of_channels_to_save, comedi_inter.sampling_rate, recording_inter.file_name);
  recording_inter.statusbar_context_id=gtk_statusbar_get_context_id(GTK_STATUSBAR(widgets.statusbar),"recording");
  recording_inter.statusbar_message_id=gtk_statusbar_push(GTK_STATUSBAR(widgets.statusbar),recording_inter.statusbar_context_id,str);
  
  // start the recording thread
  if (recording_inter.is_recording==0)
    {
      // if thread is started successfully, set the flag to 1
      if( (recording_thread_id=pthread_create(&recording_thread, NULL, recording,NULL)) ==1)
	{
	  fprintf(stderr,"error creating the recording thread, error_no:%d\n",
		  recording_thread_id);
	  return;
	}
      recording_inter.is_recording=1;
      nanosleep(&recording_inter.inter_recording_sleep_timespec,&recording_inter.req);
      // sleep to give time to open the file before starting the acquisition thread
    }
  else
    {
      fprintf(stderr,"on_record_toolbutton_clicked, recording thread was already running, this is a bug\n");
    }
  

  // start the acquisition thread
  if( (acquisition_thread_id=pthread_create(&acquisition_thread, NULL, acquisition, (void*)t))==1)
    {
      fprintf(stderr,"error creating the acquisition thread, error_no:%d\n",
	      acquisition_thread_id);
      return;
    }
  comedi_inter.is_acquiring=1;
  
}

void on_gain_increase_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
    fprintf(stderr,"gain increase button\n");
}
void on_gain_decrease_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"gain decrease button\n");
}
void on_time_increase_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
    fprintf(stderr,"time increase button\n");
}
void on_time_decrease_toolbutton_clicked(GtkObject *object, gpointer user_data)
{
  fprintf(stderr,"time decrease button\n");
}

void * recording(void * thread_id)
{
  void * pt=NULL;
  fprintf(stderr,"at top of recording() function\n");
 
  while(recording_inter.is_recording==1)
    {
      if(recording_interface_get_data(&recording_inter, &comedi_inter)!=0)
	{
	  fprintf(stderr,"recording(): problem getting new data for recording, recording aborted\n");
	  recording_inter.is_recording=0;
	  nanosleep(&recording_inter.inter_recording_sleep_timespec,&recording_inter.req);
	  recording_interface_close_file(&recording_inter);
	  recording_interface_clear_current_recording_variables(&recording_inter);
	  gtk_statusbar_remove(GTK_STATUSBAR(widgets.statusbar),recording_inter.statusbar_context_id,recording_inter.statusbar_message_id);
	  return pt;
	}
      if(recording_interface_save_data(&recording_inter)!=0)
	{
	  fprintf(stderr,"recording(): problem saving data, recording aborted\n");
	  recording_inter.is_recording=0;
	  nanosleep(&recording_inter.inter_recording_sleep_timespec,&recording_inter.req);
	  recording_interface_close_file(&recording_inter);
	  recording_interface_clear_current_recording_variables(&recording_inter);
	  gtk_statusbar_remove(GTK_STATUSBAR(widgets.statusbar),recording_inter.statusbar_context_id,recording_inter.statusbar_message_id);
	  return pt;
	}
      // sleep a bit between getting data so that you get a decent amount each time
      nanosleep(&recording_inter.inter_recording_sleep_timespec,&recording_inter.req);
    }
  fprintf(stderr,"exiting while(is_recording==1) loop\n");
  
  if(recording_interface_check_file_size(&recording_inter)!=0)
    {
      fprintf(stderr,"recording(): problem with save of the recorded file\n");
    }
  recording_interface_close_file(&recording_inter);
  recording_interface_clear_current_recording_variables(&recording_inter);
  gtk_statusbar_remove(GTK_STATUSBAR(widgets.statusbar),recording_inter.statusbar_context_id,recording_inter.statusbar_message_id);

  // increament the file index
  // get the file name from the gui
  int index=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widgets.trial_spinbutton));
  index++;
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.trial_spinbutton),(gdouble)index);


  pthread_exit(NULL);
  return pt;
}
void * acquisition(void* thread_id)
{
  void *pt=NULL;
  fprintf(stderr,"at top of acquisition() function\n");
  // start the acquisition on the comedi devices
  if (comedi_inter.command_running==0)
    {
      if(comedi_interface_run_command()!=0)
	{
	  comedi_inter.is_acquiring=0;
	  fprintf(stderr,"error returned by comedi_interface_run_command(), acquisition not possible\n");
	  return pt ;
	}
      comedi_inter.command_running=1;
    }
  while(comedi_inter.is_acquiring==1)
    {
      comedi_interface_get_data_from_devices();
      nanosleep(&comedi_inter.inter_acquisition_sleep_timespec,&comedi_inter.req);
    }
  fprintf(stderr,"exiting while(is_acquiring==1) loop, about to stop comedi command\n");
  // function to stop the comedi command from getting new data
  if(comedi_interface_stop_command()!=0)
    {
      fprintf(stderr,"error returned by comedi_interface_stop_command()\n");
      return pt ;
    }
    // function to reset the counter and variables of interface and devices
  comedi_interface_clear_current_acquisition_variables();
  
  // exiting from this function will kill the recording_thread
  pthread_exit(NULL);
  return pt;
}

struct timespec set_timespec_from_ms(double milisec)
{ // set the values in timespec structure
  struct timespec temp;
  time_t sec=(int)(milisec/1000);
  milisec=milisec-(sec*1000);
  temp.tv_sec=sec;
  temp.tv_nsec=milisec*1000000L;
  return temp;
}
struct timespec diff(struct timespec start, struct timespec end)
{
  // get the time difference between two times
  struct timespec temp;
  if ((end.tv_nsec-start.tv_nsec)<0) {
    temp.tv_sec = end.tv_sec-start.tv_sec-1;
    temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
  } else {
    temp.tv_sec = end.tv_sec-start.tv_sec;
    temp.tv_nsec = end.tv_nsec-start.tv_nsec;
  }
  return temp;
}

int microsecond_from_timespec(struct timespec duration)
{
  int ms;
  ms=duration.tv_nsec/1000;
  //  ms=ms+duration.tv_sec*1000;
  return ms;
}
